#!/usr/bin/env python
############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import commands, sys, time, urlparse, optparse, base64
import ZSI
from ZSI.wstools.Namespaces import WSA200403
from ZSI.TC import _get_type_definition, _get_global_element_declaration,\
    b64decode, b64encode, ElementDeclaration, String

from twisted.python import log
from twisted.internet import reactor

from pyGlobus import ioc
from pyGlobus.io import GSIHTTPConnection, TCPIOAttr, AuthData
from pyGlobus.security import GSICredentialAttr, GSICredential, \
    STACK_X509_Factory, STACK_X509

from pyGridWare.utility.scripts.client import GetOptParser, GetPortKWArgs, \
    SetUp, GetConfigParser
import pyGridWare.generated.stubs.managedjobservice as MJS
import pyGridWare.generated.properties.managedjobservice.ManagedJobService.ManagedJobPortTypePort as MJS_PR
import pyGridWare.generated.types.managedjobservice as MJS_TYPES
import pyGridWare.generated.stubs.managedjobfactory as MJF
import pyGridWare.generated.stubs.delegationservice as DELEGATION
import pyGridWare.generated.types.delegationservice as DELEGATION_TYPES
import pyGridWare.generated.types.managedjobfactory as MJF_TYPES
from pyGridWare.notification.NotificationConsumer import \
    NotificationConsumer, NCServiceInterface



GED = _get_global_element_declaration
class QName(tuple):
    typecode = ZSI.TC.QName()

#######################################################################
# Runtime Patch
#######################################################################
for what in MJF.createManagedJobInputMessage.typecode.ofwhat:
    """Runtime Patch for Choice ModelGroup
    """
    if what.pname == 'multiJob':
        what.minOccurs = 0

#######################################################################
# Replace twisted client w/ ZSI client that can do GSI.
#######################################################################
class MJFPortType(MJF.ManagedJobFactoryPortTypeSOAPBindingSOAP):
    def __init__(self, url, endPointReference=None, **kw):
        kw.setdefault("readerclass", MJF.DomletteReader)
        kw.setdefault("writerclass", MJF.DomletteElementProxy)
        kw.setdefault("ResourceProperties", ("http://www.globus.org/namespaces/2004/10/gram/job","managedJobFactoryResourceProperties"))
        self.binding = ZSI.client.Binding(url=url, **kw)
        self.endPointReference = endPointReference

class DFPortType(DELEGATION.DelegationFactoryPortTypeSOAPBindingSOAP):
    def __init__(self, url, endPointReference=None, **kw):
        kw.setdefault("readerclass", DELEGATION.DomletteReader)
        kw.setdefault("writerclass", DELEGATION.DomletteElementProxy)
        kw.setdefault("ResourceProperties", ("http://www.globus.org/08/2004/delegationService","RPSet"))
        self.binding = ZSI.client.Binding(url=url, **kw)
        self.endPointReference = endPointReference


#######################################################################
# NotificationHandler
#######################################################################
class NotificationHandler(NCServiceInterface):

    def __init__(self, post='/wsrf/services/NotificationConsumerService', **kw):
        NCServiceInterface.__init__(self, post)

    def wsa_Notify(self, ps, address):
        response = NCServiceInterface.wsa_Notify(self, ps, address)

        log.msg("Notification: ", self.request)
        pyobj = self.request._NotificationMessage[0]
        log.msg("  NotificationMessage: ", pyobj)
        log.msg("    Topic: ", pyobj._Topic)
        log.msg("    ProducerReference: ", pyobj._ProducerReference)
        log.msg("    Message: ", pyobj._Message)


#######################################################################
# Functions
#
#######################################################################

def CreateManagedJob(mjfEPR, cEPR, delEPR=None, *args, **kw):
    '''Essentially copying an EPR obtained from delegation serivce over
    to jobCredentialEndpoint and staggingCredentialEndpoint.

    mjfEPR -- ManagedJobFactory EPR
    cEPR   -- NotificationConsumer EPR
    delEPR -- DelegationService EPR (optional)

    '''
    # -------------------------------------------
    # Get Port to ManangedJobFactoryService
    # -------------------------------------------
    mPort = MJFPortType(endPointReference=mjfEPR, **kw)

    request = MJF.createManagedJobInputMessage()
    for what in request.typecode.ofwhat:
        if getattr(what, 'pyclass', None) is None: continue
        setattr(request, what.aname, what.pyclass())
        if what.pname == 'job':
            for j in getattr(request, what.aname).typecode.ofwhat:
                if getattr(j, 'pyclass', None) is None: continue
                setattr(getattr(request,what.aname), j.aname, j.pyclass())
                 
    class Holder: pass
    request._InitialTerminationTime = time.gmtime(time.time()+10000)

    # WARNING: UNIX command, Windows??
    request._JobID = 'uuid:%s' %commands.getoutput('uuidgen')

    # -------------------------------------------
    # Subscribing to a RP of the resource to be created.
    # ManagedJobPortType -- Resource Properties: 
    #     GED ("http://www.globus.org/namespaces/2004/10/gram/job","managedJobResourceProperties")
    #        element ref ("http://www.globus.org/namespaces/2004/10/gram/job/types","state")
    # -------------------------------------------
    request._Subscribe = Holder()
    request._Subscribe._ConsumerReference = cEPR
    request._Subscribe._TopicExpression = Holder()
    request._Subscribe._TopicExpression._text =\
        QName(('http://www.globus.org/namespaces/2004/10/gram/job/types','state'))
    request._Subscribe._TopicExpression._attrs = attrs = {}
    attrs[('http://docs.oasis-open.org/wsn/2004/06/wsn-WS-BaseNotification-1.2-draft-01.xsd','Dialect')] =\
          'http://docs.oasis-open.org/wsn/2004/06/TopicExpression/Simple'

    request._multiJob = None

    # set up job factoryEndpoint
    request._job._factoryEndpoint._Address = mjfEPR._Address
    request._job._factoryEndpoint._ReferenceProperties = Holder()
    request._job._factoryEndpoint._ReferenceProperties._any = \
        mjfEPR._ReferenceProperties._any

    if delEPR is not None:
        # set up job jobCredentialEndpoint: DelegationService
        request._job._jobCredentialEndpoint._Address = delEPR._Address
        request._job._jobCredentialEndpoint._ReferenceProperties = Holder()
        request._job._jobCredentialEndpoint._ReferenceProperties._any = \
            delEPR._ReferenceProperties._any

        # set up job stagingCredentialEndpoint: Delegation Service
        request._job._stagingCredentialEndpoint._Address = delEPR._Address
        request._job._stagingCredentialEndpoint._ReferenceProperties = Holder()
        request._job._stagingCredentialEndpoint._ReferenceProperties._any = \
            delEPR._ReferenceProperties._any
    else:
        request._job._jobCredentialEndpoint = None
        request._job._stagingCredentialEndpoint = None

    # job description
    if len(args):
        request._job._executable = args[0]
        request._job._argument = args[1:]

    request._job._localUserId = None
    request._job._holdState = None 
    request._job._directory = None
    request._job._environment = None
    request._job._stdin = None
    request._job._stdout = None
    request._job._stderr = None
    request._job._count = None
    request._job._libraryPath = None
    request._job._hostCount = None
    request._job._project = None
    request._job._queue = None
    request._job._maxTime = None
    request._job._maxWallTime = None
    request._job._maxCpuTime = None
    request._job._maxMemory = None
    request._job._minMemory = None
    request._job._jobType = None
    request._job._fileStageIn = None
    request._job._fileStageOut = None
    request._job._fileCleanUp = None
    request._job._extensions = None

    return mPort.createManagedJob(request)


def DelegateCredential(**kw):
    """Returns an EPR to the Delegated Resource
    """
    # -------------------------------------------
    # Get CertificatChain from DelegationFactoryService
    # -------------------------------------------
    kw['endPointReference'] = None
    dPort = DFPortType(**kw)
    property = ("http://www.globus.org/08/2004/delegationService","CertificateChain")
    getReq = DELEGATION.GetResourcePropertyRequest(QName(property))
    getRsp = dPort.GetResourceProperty(getReq)

    # -------------------------------------------
    # Grab BinarySecurityToken, Create X509 Stack (PKIPath)
    # -------------------------------------------
    certChain = getRsp._any[0]
    bstoken = certChain._any[0]
    assert isinstance(certChain.typecode, \
        GED('http://www.globus.org/08/2004/delegationService', 'CertificateChain')), \
        'Expecting CertificateChain'
    #assert isinstance(bstoken.typecode, \
    #    GED('http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd', 
    #        'BinarySecurityToken')), \
    #    'Expecting BinarySecurityToken'

    stackFact = STACK_X509_Factory()
    stack = stackFact.createStackFromEncodedSequence(bstoken)
    assert isinstance(stack, STACK_X509), 'stack is wrong type'

    # -------------------------------------------
    # Grab Credential, use to sign 1st cred in PKIPath .
    #    FIX: Currently just inserting credential into chain
    # -------------------------------------------
    cred = GSICredential()
    cred.read()
    publicKey = cred.get_cert()
    privateKey = cred.get_key()
    credCertChain = cred.get_cert_chain()
    assert isinstance(credCertChain, STACK_X509), 'credCertChain is wrong type'
    credCertChain.insert(publicKey, 0)

    # -------------------------------------------
    # RequestSecurityToken
    # -------------------------------------------
    bst = GED("http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd", 
              "BinarySecurityToken")().pyclass(base64.encodestring(credCertChain.get_der_encoding()))
    bst._attrs = {}
    bst._attrs['EncodingType'] =\
        "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-soap-message-security-1.0#Base64Binary"
    bst._attrs['ValueType'] =\
         "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-x509-token-profile-1.0#X509PKIPathv1"
    rstokMsg = DELEGATION.RequestSecurityTokenMsg()
    rstokMsg._any = [bst,]
    rstokMsgR = dPort.RequestSecurityToken(rstokMsg)
    
    # -------------------------------------------
    # EPR of delegated credential.
    # -------------------------------------------
    epr = rstokMsgR._any[0]
    assert isinstance(epr.typecode, \
        GED('http://schemas.xmlsoap.org/ws/2004/03/addressing','EndpointReference')), \
        'Expecting a EndpointReference'

    return epr

DELEGATION_OPTIONS = ['none', 'full', 'limited']
OPERATION_OPTIONS = ['validate', 'submit', 'monitor', 'status', 'kill']

def main(options):
    '''Using GSIHTTPConnection and ZSI.client.Binding until openssl-0.9.8 is 
    available.
    '''
    kw = {}
    mjfURL = options.url
    kw['wsAddressURI'] = WSA200403.ADDRESS
    kw['tracefile'] = sys.stdout
    kw['transport'] = GSIHTTPConnection
    kw['transdict'] = {\
        'channel_mode':ioc.GLOBUS_IO_SECURE_CHANNEL_MODE_SSL_WRAP,
        'delegation_mode':ioc.GLOBUS_IO_SECURE_DELEGATION_MODE_NONE,
    }
    WSGRAM = 'http://www.globus.org/namespaces/2004/10/gram/job'

    # -------------------------------------------
    # Create EPR for "Fork"
    # -------------------------------------------
    mjfEPR = _get_global_element_declaration(WSA200403.ADDRESS, 'EndpointReference')().pyclass()
    mjfEPR._ReferenceProperties = \
        filter(lambda what: what.type==(WSA200403.ADDRESS,"ReferencePropertiesType"),
               mjfEPR.typecode.ofwhat)[0].pyclass()
               
    class ResourceID(str):
        typecode = String((WSGRAM,"ResourceID"), typed=False)
    mjfEPR._ReferenceProperties._any = [ResourceID('Fork')]
    mjfEPR._Address = mjfURL

    # -------------------------------------------
    # Create NotificationConsumer and get EPR
    # -------------------------------------------
    consumer = NotificationConsumer(notificationClass=NotificationHandler)
    consumer.start()
    cEPR = consumer.context.getConsumerReference()
    delEPR = None
    if options.delegate != DELEGATION_OPTIONS[0]:
        # -------------------------------------------
        # Get Port to ManangedJobFactoryService
        # -------------------------------------------
        mPort = MJFPortType(url=mjfURL, endPointReference=mjfEPR, **kw)

        # -------------------------------------------
        # Create GetMultipleResourcePropertiesRequest
        # -------------------------------------------
        multiReq = MJF.GetMultipleResourcePropertiesRequest()
        multiReq._ResourceProperty = [\
           QName((WSGRAM, 'delegationFactoryEndpoint')),
           QName((WSGRAM, 'stagingDelegationFactoryEndpoint')),
        ]
        multiRsp = mPort.GetMultipleResourceProperties(multiReq)

        # -------------------------------------------
        # Delegate Credential
        # -------------------------------------------
        scheme,netloc,url,params,query,fragment = urlparse.urlparse(mjfURL)
        params,query,fragment = '','',''
        cp = GetConfigParser()
        path = cp.get('Paths', 'DelegationFactoryService')
        dcURL = urlparse.urlunparse((scheme,netloc,path,params,query,fragment))
        delEPR = DelegateCredential(url=dcURL, **kw)

    # -------------------------------------------
    # Create Manged Job
    # -------------------------------------------
    CreateManagedJob(mjfEPR, cEPR, delEPR, url=mjfURL, *args, **kw)


if __name__ == '__main__':
    parser = GetOptParser('ManagedJobFactoryService')
    parser.add_option('-t', "--delegate", action="store", dest="delegate", 
        choices=DELEGATION_OPTIONS, help="choose from %s" %DELEGATION_OPTIONS)
    parser.add_option('-o', "--operation", action="store", dest="operation", 
        choices=OPERATION_OPTIONS, default='submit', help="choose %s" %OPERATION_OPTIONS)
    (options, args) = parser.parse_args()
    SetUp(options)

    reactor.callWhenRunning(main, options)
    reactor.run()
